import abc
import os
import io
import random
import shutil
import sys
import tempfile
import traceback
import unittest
from contextlib import contextmanager
from datetime import datetime

import yaml
from unittest.mock import patch

import dbt.main as dbt
import dbt.flags as flags
from dbt.deprecations import reset_deprecations
from dbt.adapters.factory import get_adapter, reset_adapters, register_adapter
from dbt.clients.jinja import template_cache
from dbt.config.runtime import RuntimeConfig
from dbt.context import providers
from dbt.logger import GLOBAL_LOGGER as logger, log_manager


INITIAL_ROOT = os.getcwd()


def normalize(path):
    """On windows, neither is enough on its own:

    >>> normcase('C:\\documents/ALL CAPS/subdir\\..')
    'c:\\documents\\all caps\\subdir\\..'
    >>> normpath('C:\\documents/ALL CAPS/subdir\\..')
    'C:\\documents\\ALL CAPS'
    >>> normpath(normcase('C:\\documents/ALL CAPS/subdir\\..'))
    'c:\\documents\\all caps'
    """
    return os.path.normcase(os.path.normpath(path))


class TestArgs:
    def __init__(self, kwargs):
        self.which = 'run'
        self.single_threaded = False
        self.profiles_dir = None
        self.project_dir = None
        self.__dict__.update(kwargs)


def _pytest_test_name():
    return os.environ['PYTEST_CURRENT_TEST'].split()[0]


def _pytest_get_test_root():
    test_path = _pytest_test_name().split('::')[0]
    relative_to = INITIAL_ROOT
    head = os.path.relpath(test_path, relative_to)

    path_parts = []
    while head:
        head, tail = os.path.split(head)
        path_parts.append(tail)
    path_parts.reverse()
    # dbt tests are all of the form 'test/integration/XXX_suite_name'
    target = os.path.join(*path_parts[:3])
    return os.path.join(relative_to, target)


def _really_makedirs(path):
    while not os.path.exists(path):
        try:
            os.makedirs(path)
        except EnvironmentError:
            raise


class DBTIntegrationTestBase(unittest.TestCase, metaclass=abc.ABCMeta):
    CREATE_SCHEMA_STATEMENT = 'CREATE SCHEMA {}'
    DROP_SCHEMA_STATEMENT = 'DROP SCHEMA IF EXISTS {} CASCADE'

    _randint = random.randint(0, 9999)
    _runtime_timedelta = (datetime.utcnow() - datetime(1970, 1, 1, 0, 0, 0))
    _runtime = (
        (int(_runtime_timedelta.total_seconds() * 1e6)) +
        _runtime_timedelta.microseconds
    )

    prefix = f'test{_runtime}{_randint:04}'

    @property
    def schema(self):
        raise NotImplementedError

    @property
    def models(self):
        raise NotImplementedError

    @property
    def database_host(self):
        if os.name == 'nt':
            return 'localhost'
        return 'database'

    @property
    def packages_config(self):
        return None

    @property
    def selectors_config(self):
        return None

    def unique_schema(self):
        schema = self.schema

        to_return = "{}_{}".format(self.prefix, schema)

        return to_return.lower()

    @property
    def default_database(self):
        database = self.config.credentials.database
        return database

    @staticmethod
    @abc.abstractmethod
    def _profile_from_test_name(test_name):
        pass

    def _pick_profile(self):
        test_name = self.id().split('.')[-1]
        return self._profile_from_test_name(test_name)

    def _symlink_test_folders(self):
        for entry in os.listdir(self.test_original_source_path):
            src = os.path.join(self.test_original_source_path, entry)
            tst = os.path.join(self.test_root_dir, entry)
            if os.path.isdir(src) or src.endswith('.sql'):
                # symlink all sql files and all directories.
                os.symlink(src, tst)
        os.symlink(self._logs_dir, os.path.join(self.test_root_dir, 'logs'))

    @property
    def test_root_realpath(self):
        if sys.platform == 'darwin':
            return os.path.realpath(self.test_root_dir)
        else:
            return self.test_root_dir

    def _generate_test_root_dir(self):
        return normalize(tempfile.mkdtemp(prefix='dbt-int-test-'))

    def setUp(self):
        self.dbt_core_install_root = os.path.dirname(dbt.__file__)
        log_manager.reset_handlers()
        self.initial_dir = INITIAL_ROOT
        os.chdir(self.initial_dir)
        # before we go anywhere, collect the initial path info
        self._logs_dir = os.path.join(self.initial_dir, 'logs', self.prefix)
        _really_makedirs(self._logs_dir)
        self.test_original_source_path = _pytest_get_test_root()
        self.test_root_dir = self._generate_test_root_dir()

        os.chdir(self.test_root_dir)
        try:
            self._symlink_test_folders()
        except Exception:
            msg = '\n\t'.join((
                'Failed to symlink test folders!',
                'initial_dir={0.initial_dir}',
                'test_original_source_path={0.test_original_source_path}',
                'test_root_dir={0.test_root_dir}'
            )).format(self)
            logger.exception(msg)

            # if logging isn't set up, I still really want this message.
            print(msg)
            traceback.print_exc()

            raise

        self._created_schemas = set()
        reset_deprecations()
        flags.reset()
        template_cache.clear()

        self.use_profile(self._pick_profile())
        self.use_default_project()
        self.set_packages()
        self.set_selectors()
        self.load_config()

    def use_default_project(self, overrides=None):
        # create a dbt_project.yml
        base_project_config = {
            'name': 'test',
            'version': '1.0',
            'config-version': 2,
            'test-paths': [],
            'source-paths': [self.models],
            'profile': 'test',
        }

        project_config = {}
        project_config.update(base_project_config)
        project_config.update(self.project_config)
        project_config.update(overrides or {})

        with open("dbt_project.yml", 'w') as f:
            yaml.safe_dump(project_config, f, default_flow_style=True)

    @abc.abstractmethod
    def get_profile(self, adapter_type):
        pass

    def use_profile(self, adapter_type):
        self.adapter_type = adapter_type

        profile_config = {}
        default_profile_config = self.get_profile(adapter_type)

        profile_config.update(default_profile_config)
        profile_config.update(self.profile_config)

        if not os.path.exists(self.test_root_dir):
            os.makedirs(self.test_root_dir)

        profiles_path = os.path.join(self.test_root_dir, 'profiles.yml')
        with open(profiles_path, 'w') as f:
            yaml.safe_dump(profile_config, f, default_flow_style=True)
        self._profile_config = profile_config

    def set_packages(self):
        if self.packages_config is not None:
            with open('packages.yml', 'w') as f:
                yaml.safe_dump(self.packages_config, f,
                               default_flow_style=True)

    def set_selectors(self):
        if self.selectors_config is not None:
            with open('selectors.yml', 'w') as f:
                yaml.safe_dump(self.selectors_config, f,
                               default_flow_style=True)

    def load_config(self):
        # we've written our profile and project. Now we want to instantiate a
        # fresh adapter for the tests.
        # it's important to use a different connection handle here so
        # we don't look into an incomplete transaction
        kwargs = {
            'profile': None,
            'profiles_dir': self.test_root_dir,
            'target': None,
        }

        config = RuntimeConfig.from_args(TestArgs(kwargs))

        register_adapter(config)
        adapter = get_adapter(config)
        adapter.cleanup_connections()
        self.adapter_type = adapter.type()
        self.adapter = adapter
        self.config = config

        self._drop_schemas()
        self._create_schemas()

    def quote_as_configured(self, value, quote_key):
        return self.adapter.quote_as_configured(value, quote_key)

    def tearDown(self):
        # get any current run adapter and clean up its connections before we
        # reset them. It'll probably be different from ours because
        # handle_and_check() calls reset_adapters().
        register_adapter(self.config)
        adapter = get_adapter(self.config)
        if adapter is not self.adapter:
            adapter.cleanup_connections()
        if not hasattr(self, 'adapter'):
            self.adapter = adapter

        self._drop_schemas()

        self.adapter.cleanup_connections()
        reset_adapters()
        os.chdir(INITIAL_ROOT)
        try:
            shutil.rmtree(self.test_root_dir)
        except EnvironmentError:
            logger.exception('Could not clean up after test - {} not removable'
                             .format(self.test_root_dir))

    def _get_schema_fqn(self, database, schema):
        return self.quote_as_configured(schema, 'schema')

    def _create_schema_named(self, database, schema):
        schema_fqn = self._get_schema_fqn(database, schema)
        self.run_sql(self.CREATE_SCHEMA_STATEMENT.format(schema_fqn))
        self._created_schemas.add(schema_fqn)

    def _create_schemas(self):
        schema = self.unique_schema()
        with self.adapter.connection_named('__test'):
            self._create_schema_named(self.default_database, schema)

    def _drop_schemas_sql(self):
        schema = self.unique_schema()
        # we always want to drop these if necessary, we'll clear it soon.
        self._created_schemas.add(
            self._get_schema_fqn(self.default_database, schema)
        )

        for schema_fqn in self._created_schemas:
            self.run_sql(self.DROP_SCHEMA_STATEMENT.format(schema_fqn))

        self._created_schemas.clear()

    def _drop_schemas(self):
        with self.adapter.connection_named('__test'):
            self._drop_schemas_sql()

    @property
    def project_config(self):
        return {
            'config-version': 2,
        }

    @property
    def profile_config(self):
        return {}

    def run_dbt(self, args=None, expect_pass=True, strict=True, parser=True, profiles_dir=True):
        res, success = self.run_dbt_and_check(
            args=args, strict=strict, parser=parser, profiles_dir=profiles_dir)

        self.assertEqual(
            success, expect_pass,
            "dbt exit state did not match expected")

        return res

    def run_dbt_and_capture(self, *args, **kwargs):
        try:
            initial_stdout = log_manager.stdout
            initial_stderr = log_manager.stderr
            stringbuf = io.StringIO()
            log_manager.set_output_stream(stringbuf)

            res = self.run_dbt(*args, **kwargs)
            stdout = stringbuf.getvalue()

        finally:
            log_manager.set_output_stream(initial_stdout, initial_stderr)

        return res, stdout

    def run_dbt_and_check(self, args=None, strict=True, parser=False, profiles_dir=True):
        log_manager.reset_handlers()
        if args is None:
            args = ["run"]

        final_args = []

        if strict:
            final_args.append('--strict')
        if parser:
            final_args.append('--test-new-parser')
        if os.getenv('DBT_TEST_SINGLE_THREADED') in ('y', 'Y', '1'):
            final_args.append('--single-threaded')

        final_args.extend(args)

        if profiles_dir:
            final_args.extend(['--profiles-dir', self.test_root_dir])
        final_args.append('--log-cache-events')

        logger.info("Invoking dbt with {}".format(final_args))
        return dbt.handle_and_check(final_args)

    def run_sql_file(self, path, kwargs=None):
        with open(path, 'r') as f:
            statements = f.read().split(";")
            for statement in statements:
                self.run_sql(statement, kwargs=kwargs)

    def transform_sql(self, query, kwargs=None):
        to_return = query
        base_kwargs = {
            'schema': self.unique_schema(),
            'database': self.adapter.quote(self.default_database),
        }
        if kwargs is None:
            kwargs = {}
        base_kwargs.update(kwargs)

        to_return = to_return.format(**base_kwargs)

        return to_return

    def run_sql_common(self, sql, fetch, conn):
        with conn.handle.cursor() as cursor:
            try:
                cursor.execute(sql)
                conn.handle.commit()
                if fetch == 'one':
                    return cursor.fetchone()
                elif fetch == 'all':
                    return cursor.fetchall()
                else:
                    return
            except BaseException as e:
                if conn.handle and not getattr(conn.handle, 'closed', True):
                    conn.handle.rollback()
                print(sql)
                print(e)
                raise
            finally:
                conn.transaction_open = False

    def run_sql(self, query, fetch='None', kwargs=None, connection_name=None):
        if connection_name is None:
            connection_name = '__test'

        if query.strip() == "":
            return

        sql = self.transform_sql(query, kwargs=kwargs)

        with self.get_connection(connection_name) as conn:
            logger.debug(
                'test connection "{}" executing: {}'.format(conn.name, sql))
            return self.run_sql_common(sql, fetch, conn)

    def _ilike(self, target, value):
        return "{} ilike '{}'".format(target, value)

    def get_many_table_columns_information_schema(self, tables, schema, database=None):
        columns = 'table_name, column_name, data_type, character_maximum_length'

        sql = """
                select {columns}
                from {db_string}information_schema.columns
                where {schema_filter}
                  and ({table_filter})
                order by column_name asc"""

        db_string = ''
        if database:
            db_string = self.quote_as_configured(database, 'database') + '.'

        table_filters_s = " OR ".join(
            self._ilike('table_name', table.replace('"', ''))
            for table in tables
        )
        schema_filter = self._ilike('table_schema', schema)

        sql = sql.format(
            columns=columns,
            schema_filter=schema_filter,
            table_filter=table_filters_s,
            db_string=db_string)

        columns = self.run_sql(sql, fetch='all')
        return list(map(self.filter_many_columns, columns))

    def get_many_table_columns(self, tables, schema, database=None):
        result = self.get_many_table_columns_information_schema(
            tables, schema, database)
        result.sort(key=lambda x: '{}.{}'.format(x[0], x[1]))
        return result

    def filter_many_columns(self, column):
        if len(column) == 3:
            table_name, column_name, data_type = column
            char_size = None
        else:
            table_name, column_name, data_type, char_size = column
        return (table_name, column_name, data_type, char_size)

    @contextmanager
    def get_connection(self, name=None):
        """Create a test connection context where all executed macros, etc will
        get self.adapter as the adapter.

        This allows tests to run normal adapter macros as if reset_adapters()
        were not called by handle_and_check (for asserts, etc)
        """
        if name is None:
            name = '__test'
        with patch.object(providers, 'get_adapter', return_value=self.adapter):
            with self.adapter.connection_named(name):
                conn = self.adapter.connections.get_thread_connection()
                yield conn

    def get_relation_columns(self, relation):
        with self.get_connection():
            columns = self.adapter.get_columns_in_relation(relation)

        return sorted(((c.name, c.dtype, c.char_size) for c in columns),
                      key=lambda x: x[0])

    def get_table_columns(self, table, schema=None, database=None):
        schema = self.unique_schema() if schema is None else schema
        database = self.default_database if database is None else database
        relation = self.adapter.Relation.create(
            database=database,
            schema=schema,
            identifier=table,
            type='table',
            quote_policy=self.config.quoting
        )
        return self.get_relation_columns(relation)

    def get_table_columns_as_dict(self, tables, schema=None):
        col_matrix = self.get_many_table_columns(tables, schema)
        res = {}
        for row in col_matrix:
            table_name = row[0]
            col_def = row[1:]
            if table_name not in res:
                res[table_name] = []
            res[table_name].append(col_def)
        return res

    def get_models_in_schema(self, schema=None):
        schema = self.unique_schema() if schema is None else schema

        sql = """
                select table_name,
                        case when table_type = 'BASE TABLE' then 'table'
                             when table_type = 'VIEW' then 'view'
                             else table_type
                        end as materialization
                from information_schema.tables
                where {}
                order by table_name
                """

        sql = sql.format(self._ilike('table_schema', schema))
        result = self.run_sql(sql, fetch='all')

        return {model_name: materialization for (model_name, materialization) in result}

    def _assertTablesEqualSql(self, relation_a, relation_b, columns=None):
        if columns is None:
            columns = self.get_relation_columns(relation_a)
        column_names = [c[0] for c in columns]

        sql = self.adapter.get_rows_different_sql(
            relation_a, relation_b, column_names
        )

        return sql

    def assertTablesEqual(self, table_a, table_b,
                          table_a_schema=None, table_b_schema=None,
                          table_a_db=None, table_b_db=None):
        if table_a_schema is None:
            table_a_schema = self.unique_schema()

        if table_b_schema is None:
            table_b_schema = self.unique_schema()

        if table_a_db is None:
            table_a_db = self.default_database

        if table_b_db is None:
            table_b_db = self.default_database

        relation_a = self._make_relation(table_a, table_a_schema, table_a_db)
        relation_b = self._make_relation(table_b, table_b_schema, table_b_db)

        self._assertTableColumnsEqual(relation_a, relation_b)

        sql = self._assertTablesEqualSql(relation_a, relation_b)
        result = self.run_sql(sql, fetch='one')

        self.assertEqual(
            result[0],
            0,
            'row_count_difference nonzero: ' + sql
        )
        self.assertEqual(
            result[1],
            0,
            'num_mismatched nonzero: ' + sql
        )

    def _make_relation(self, identifier, schema=None, database=None):
        if schema is None:
            schema = self.unique_schema()
        if database is None:
            database = self.default_database
        return self.adapter.Relation.create(
            database=database,
            schema=schema,
            identifier=identifier,
            quote_policy=self.config.quoting
        )

    def get_many_relation_columns(self, relations):
        """Returns a dict of (datbase, schema) -> (dict of (table_name -> list of columns))
        """
        schema_fqns = {}
        for rel in relations:
            this_schema = schema_fqns.setdefault(
                (rel.database, rel.schema), [])
            this_schema.append(rel.identifier)

        column_specs = {}
        for key, tables in schema_fqns.items():
            database, schema = key
            columns = self.get_many_table_columns(
                tables, schema, database=database)
            table_columns = {}
            for col in columns:
                table_columns.setdefault(col[0], []).append(col[1:])
            for rel_name, columns in table_columns.items():
                key = (database, schema, rel_name)
                column_specs[key] = columns

        return column_specs

    def assertManyRelationsEqual(self, relations, default_schema=None, default_database=None):
        if default_schema is None:
            default_schema = self.unique_schema()
        if default_database is None:
            default_database = self.default_database

        specs = []
        for relation in relations:
            if not isinstance(relation, (tuple, list)):
                relation = [relation]

            assert len(relation) <= 3

            if len(relation) == 3:
                relation = self._make_relation(*relation)
            elif len(relation) == 2:
                relation = self._make_relation(
                    relation[0], relation[1], default_database)
            elif len(relation) == 1:
                relation = self._make_relation(
                    relation[0], default_schema, default_database)
            else:
                raise ValueError(
                    'relation must be a sequence of 1, 2, or 3 values')

            specs.append(relation)

        with self.get_connection():
            column_specs = self.get_many_relation_columns(specs)

        # make sure everyone has equal column definitions
        first_columns = None
        for relation in specs:
            key = (relation.database, relation.schema, relation.identifier)
            # get a good error here instead of a hard-to-diagnose KeyError
            self.assertIn(key, column_specs, f'No columns found for {key}')
            columns = column_specs[key]
            if first_columns is None:
                first_columns = columns
            else:
                self.assertEqual(
                    first_columns, columns,
                    '{} did not match {}'.format(str(specs[0]), str(relation))
                )

        # make sure everyone has the same data. if we got here, everyone had
        # the same column specs!
        first_relation = None
        for relation in specs:
            if first_relation is None:
                first_relation = relation
            else:
                sql = self._assertTablesEqualSql(first_relation, relation,
                                                 columns=first_columns)
                result = self.run_sql(sql, fetch='one')

                self.assertEqual(
                    result[0],
                    0,
                    'row_count_difference nonzero: ' + sql
                )
                self.assertEqual(
                    result[1],
                    0,
                    'num_mismatched nonzero: ' + sql
                )

    def assertManyTablesEqual(self, *args):
        schema = self.unique_schema()

        all_tables = []
        for table_equivalencies in args:
            all_tables += list(table_equivalencies)

        all_cols = self.get_table_columns_as_dict(all_tables, schema)

        for table_equivalencies in args:
            first_table = table_equivalencies[0]
            first_relation = self._make_relation(first_table)

            # assert that all tables have the same columns
            base_result = all_cols[first_table]
            self.assertTrue(len(base_result) > 0)

            for other_table in table_equivalencies[1:]:
                other_result = all_cols[other_table]
                self.assertTrue(len(other_result) > 0)
                self.assertEqual(base_result, other_result)

                other_relation = self._make_relation(other_table)
                sql = self._assertTablesEqualSql(first_relation,
                                                 other_relation,
                                                 columns=base_result)
                result = self.run_sql(sql, fetch='one')

                self.assertEqual(
                    result[0],
                    0,
                    'row_count_difference nonzero: ' + sql
                )
                self.assertEqual(
                    result[1],
                    0,
                    'num_mismatched nonzero: ' + sql
                )

    def _assertTableRowCountsEqual(self, relation_a, relation_b):
        cmp_query = """
            with table_a as (

                select count(*) as num_rows from {}

            ), table_b as (

                select count(*) as num_rows from {}

            )

            select table_a.num_rows - table_b.num_rows as difference
            from table_a, table_b

        """.format(str(relation_a), str(relation_b))

        res = self.run_sql(cmp_query, fetch='one')

        self.assertEqual(int(res[0]), 0, "Row count of table {} doesn't match row count of table {}. ({} rows different)".format(
            relation_a.identifier,
            relation_b.identifier,
            res[0]
        )
        )

    def assertTableDoesNotExist(self, table, schema=None, database=None):
        columns = self.get_table_columns(table, schema, database)

        self.assertEqual(
            len(columns),
            0
        )

    def assertTableDoesExist(self, table, schema=None, database=None):
        columns = self.get_table_columns(table, schema, database)

        self.assertGreater(
            len(columns),
            0
        )

    def _assertTableColumnsEqual(self, relation_a, relation_b):
        table_a_result = self.get_relation_columns(relation_a)
        table_b_result = self.get_relation_columns(relation_b)

        text_types = {'text', 'character varying', 'character', 'varchar'}

        self.assertEqual(len(table_a_result), len(table_b_result))
        for a_column, b_column in zip(table_a_result, table_b_result):
            a_name, a_type, a_size = a_column
            b_name, b_type, b_size = b_column
            self.assertEqual(a_name, b_name,
                             '{} vs {}: column "{}" != "{}"'.format(
                                 relation_a, relation_b, a_name, b_name
                             ))

            self.assertEqual(a_type, b_type,
                             '{} vs {}: column "{}" has type "{}" != "{}"'.format(
                                 relation_a, relation_b, a_name, a_type, b_type
                             ))

            self.assertEqual(a_size, b_size,
                             '{} vs {}: column "{}" has size "{}" != "{}"'.format(
                                 relation_a, relation_b, a_name, a_size, b_size
                             ))

    def assertBetween(self, timestr, start, end=None):
        datefmt = '%Y-%m-%dT%H:%M:%S.%fZ'
        if end is None:
            end = datetime.utcnow()

        parsed = datetime.strptime(timestr, datefmt)

        self.assertLessEqual(start, parsed,
                             'parsed date {} happened before {}'.format(
                                 parsed,
                                 start.strftime(datefmt))
                             )
        self.assertGreaterEqual(end, parsed,
                                'parsed date {} happened after {}'.format(
                                    parsed,
                                    end.strftime(datefmt))
                                )



from .spec_file import DbtSpecFile


# custom test loader
def pytest_collect_file(parent, path):
    if path.ext == ".dbtspec":  # and path.basename.startswith("test"):
        return DbtSpecFile.from_parent(parent, fspath=path)


def pytest_addoption(parser):
    group = parser.getgroup('dbtadapter')
    group.addoption(
        '--no-drop-schema',
        action='store_false',
        dest='drop_schema',

    )



import os
from pathlib import Path
from typing import Dict, Any

import yaml

from .exceptions import TestProcessingException


DEFAULT_DBT_PROJECT = {
    'name': 'dbt_test_project',
    'config-version': 2,
    'version': '1.0.0',
}


NAMES_BASE = """
id,name,some_date
1,Easton,1981-05-20T06:46:51
2,Lillian,1978-09-03T18:10:33
3,Jeremiah,1982-03-11T03:59:51
4,Nolan,1976-05-06T20:21:35
5,Hannah,1982-06-23T05:41:26
6,Eleanor,1991-08-10T23:12:21
7,Lily,1971-03-29T14:58:02
8,Jonathan,1988-02-26T02:55:24
9,Adrian,1994-02-09T13:14:23
10,Nora,1976-03-01T16:51:39
""".lstrip()


NAMES_EXTENDED = NAMES_BASE + """
11,Mateo,2014-09-07T17:04:27
12,Julian,2000-02-04T11:48:30
13,Gabriel,2001-07-10T07:32:52
14,Isaac,2002-11-24T03:22:28
15,Levi,2009-11-15T11:57:15
16,Elizabeth,2005-04-09T03:50:11
17,Grayson,2019-08-06T19:28:17
18,Dylan,2014-03-01T11:50:41
19,Jayden,2009-06-06T07:12:49
20,Luke,2003-12-05T21:42:18
""".lstrip()


NAMES_ADD_COLUMN = """
id,name,some_date,last_initial
1,Easton,1981-05-20T06:46:51,A
2,Lillian,1978-09-03T18:10:33,B
3,Jeremiah,1982-03-11T03:59:51,C
4,Nolan,1976-05-06T20:21:35,D
5,Hannah,1982-06-23T05:41:26,E
6,Eleanor,1991-08-10T23:12:21,F
7,Lily,1971-03-29T14:58:02,G
8,Jonathan,1988-02-26T02:55:24,H
9,Adrian,1994-02-09T13:14:23,I
10,Nora,1976-03-01T16:51:39,J
""".lstrip()


class Model:
    def __init__(self, config, body):
        self.config = config
        self.body = body

    @classmethod
    def from_dict(cls, dct):
        try:
            config = dct.get('config', {})
            if 'materialized' in dct:
                config['materialized'] = dct['materialized']
            return cls(config=config, body=dct['body'])
        except KeyError as exc:
            raise TestProcessingException(
                f'Invalid test, model is missing key {exc}'
            )

    def config_params(self):
        if not self.config:
            return ''
        else:
            pairs = ', '.join(
                '{!s}={!r}'.format(key, value)
                for key, value in self.config.items()
            )
            return '{{ config(' + pairs + ') }}'

    def render(self):
        return '\n'.join([self.config_params(), self.body])


class DbtProject:
    def __init__(
        self,
        name: str,
        dbt_project_yml: str,
        paths: Dict[str, str],
        facts: Dict[str, Any]
    ):
        self.name = name
        self.dbt_project_yml = dbt_project_yml
        self.paths = paths
        self.facts = facts

    def write(self, path: str):
        project_path = os.path.join(path, 'project')
        os.makedirs(project_path)
        with open(os.path.join(project_path, 'dbt_project.yml'), 'w') as fp:
            fp.write(yaml.safe_dump(self.dbt_project_yml))

        for relpath, contents in self.paths.items():
            fullpath = os.path.join(project_path, relpath)
            os.makedirs(os.path.dirname(fullpath), exist_ok=True)

            if isinstance(contents, str) and contents.startswith('files.'):
                contents_path = contents.split('.')[1:]
                cur = KNOWN_FILES
                for part in contents_path:
                    if part not in cur:
                        raise TestProcessingException(
                            f'at known file lookup {contents}, could not find '
                            f'part {part} in known files path'
                        )
                    cur = cur[part]
                contents = cur

            if relpath.startswith('models/') and relpath.endswith('.sql'):
                if isinstance(contents, dict):
                    model = Model.from_dict(contents)
                    contents = model.render()
            if not isinstance(contents, str):
                raise TestProcessingException(f'{contents} is not a string')

            with open(fullpath, 'w') as fp:
                fp.write(contents)

    @classmethod
    def from_dict(cls, dct, overriding=None):
        if overriding is None:
            overriding = {}

        paths: Dict[str, Any]
        facts: Dict[str, Any]
        dbt_project_yml: Dict[str, Any]

        if 'overrides' in dct:
            name = dct['overrides']
            if name not in overriding:
                raise TestProcessingException(
                    f'Invalid project definition, override name {name} not '
                    'known'
                ) from None
            dbt_project_yml = overriding[name].dbt_project_yml.copy()
            paths = overriding[name].paths.copy()
            facts = overriding[name].facts.copy()
        else:
            try:
                name = dct['name']
            except KeyError:
                raise TestProcessingException(
                    f'Invalid project definition, no name in {dct}'
                ) from None

            dbt_project_yml = DEFAULT_DBT_PROJECT.copy()
            paths = {}
            facts = {}

        dbt_project_yml.update(dct.get('dbt_project_yml', {}))

        paths.update(dct.get('paths', {}))
        facts.update(dct.get('facts', {}))
        return cls(
            name=name,
            dbt_project_yml=dbt_project_yml,
            paths=paths,
            facts=facts,
        )


SEED_SOURCE_YML = """
version: 2
sources:
  - name: raw
    schema: "{{ target.schema }}"
    tables:
      - name: seed
        identifier: "{{ var('seed_name', 'base') }}"
"""

TEST_SEEDS_SCHEMA_YML_TEST_BASE = """
version: 2
models:
  - name: base
    columns:
     - name: id
       tests:
         - not_null
"""

TEST_MODELS_SCHEMA_YML_TEST_VIEW = """
version: 2
models:
  - name: view
    columns:
     - name: id
       tests:
         - not_null
"""

TEST_MODELS_SCHEMA_YML_TEST_TABLE = """
version: 2
models:
  - name: table
    columns:
     - name: id
       tests:
         - not_null
"""


TEST_PASSING_DATA_TEST = """
select 1 as id where id = 2
"""

TEST_FAILING_DATA_TEST = """
select 1 as id where id = 1
"""


TEST_EPHEMERAL_DATA_TEST_PASSING = '''
with my_other_cool_cte as (
    select id, name from {{ ref('ephemeral') }}
    where id > 1000
)
select name, id from my_other_cool_cte
'''


TEST_EPHEMERAL_DATA_TEST_FAILING = '''
with my_other_cool_cte as (
    select id, name from {{ ref('ephemeral') }}
    where id < 1000
)
select name, id from my_other_cool_cte
'''

INCREMENTAL_MODEL = """
select * from {{ source('raw', 'seed') }}
{% if is_incremental() %}
where id > (select max(id) from {{ this }})
{% endif %}
""".strip()


CC_ALL_SNAPSHOT_SQL = '''
{% snapshot cc_all_snapshot %}
    {{ config(
        check_cols='all', unique_key='id', strategy='check',
        target_database=database, target_schema=schema
    ) }}
    select * from {{ ref(var('seed_name', 'base')) }}
{% endsnapshot %}
'''.strip()


CC_NAME_SNAPSHOT_SQL = '''
{% snapshot cc_name_snapshot %}
    {{ config(
        check_cols=['name'], unique_key='id', strategy='check',
        target_database=database, target_schema=schema
    ) }}
    select * from {{ ref(var('seed_name', 'base')) }}
{% endsnapshot %}
'''.strip()


CC_DATE_SNAPSHOT_SQL = '''
{% snapshot cc_date_snapshot %}
    {{ config(
        check_cols=['some_date'], unique_key='id', strategy='check',
        target_database=database, target_schema=schema
    ) }}
    select * from {{ ref(var('seed_name', 'base')) }}
{% endsnapshot %}
'''.strip()


TS_SNAPSHOT_SQL = '''
{% snapshot ts_snapshot %}
    {{ config(
        strategy='timestamp',
        unique_key='id',
        updated_at='some_date',
        target_database=database,
        target_schema=schema,
    )}}
    select * from {{ ref(var('seed_name', 'base')) }}
{% endsnapshot %}
'''.strip()


EPHEMERAL_WITH_CTE = """
with my_cool_cte as (
  select name, id from {{ ref('base') }}
)
select id, name from my_cool_cte where id is not null
"""


KNOWN_FILES = {
    'seeds': {
        'base': NAMES_BASE,
        'newcolumns': NAMES_ADD_COLUMN,
        'added': NAMES_EXTENDED,
    },
    'models': {
        'base_materialized_var': """
            {{ config(materialized=var("materialized_var", "table"))}}
            select * from {{ source('raw', 'seed') }}
        """,
        'base_table': {
            'materialized': 'table',
            'body': "select * from {{ source('raw', 'seed') }}",
        },
        'base_view': {
            'materialized': 'view',
            'body': "select * from {{ source('raw', 'seed') }}",
        },
        'ephemeral': {
            'materialized': 'ephemeral',
            'body': "select * from {{ source('raw', 'seed') }}",
        },
        'ephemeral_with_cte': {
            'materialized': 'ephemeral',
            'body': EPHEMERAL_WITH_CTE,
        },
        'ephemeral_view': {
            'materialized': 'view',
            'body': "select * from {{ ref('ephemeral') }}",
        },
        'ephemeral_table': {
            'materialized': 'table',
            'body': "select * from {{ ref('ephemeral') }}",
        },
        'incremental': {
            'materialized': 'incremental',
            'body': INCREMENTAL_MODEL,
        }
    },
    'snapshots': {
        'check_cols_all': CC_ALL_SNAPSHOT_SQL,
        'check_cols_name': CC_NAME_SNAPSHOT_SQL,
        'check_cols_date': CC_DATE_SNAPSHOT_SQL,
        'timestamp': TS_SNAPSHOT_SQL,
    },
    'tests': {
        'passing': TEST_PASSING_DATA_TEST,
        'failing': TEST_FAILING_DATA_TEST,
        'ephemeral': {
            'passing': TEST_EPHEMERAL_DATA_TEST_PASSING,
            'failing': TEST_EPHEMERAL_DATA_TEST_FAILING,
        }
    },
    'schemas': {
        'base': SEED_SOURCE_YML,
        'test_seed': TEST_SEEDS_SCHEMA_YML_TEST_BASE,
        'test_view': TEST_MODELS_SCHEMA_YML_TEST_VIEW,
        'test_table': TEST_MODELS_SCHEMA_YML_TEST_TABLE,
    },
}


THIS_DIR = Path(__file__).parent
PROJECT_DIR = THIS_DIR / 'projects'
SEQUENCE_DIR = THIS_DIR / 'sequences'


def _get_named_yaml_dicts(path: Path) -> Dict[str, Dict[str, Any]]:
    result = {}
    for project_path in path.glob('**/*.yml'):
        try:
            data = yaml.safe_load(project_path.read_text())
            name = data['name']
        except KeyError as exc:
            raise ImportError(
                f'Invalid project file at {project_path}: no name'
            ) from exc
        except Exception as exc:
            raise ImportError(
                f'Could not read project file at {project_path}: {exc}'
            ) from exc
        result[name] = data
    return result


DEFAULT_PROJECTS = _get_named_yaml_dicts(PROJECT_DIR)

BUILTIN_TEST_SEQUENCES = _get_named_yaml_dicts(SEQUENCE_DIR)



import json
import os
import random
import shlex
import tempfile
from datetime import datetime
from itertools import chain, repeat
from subprocess import run, CalledProcessError, PIPE
from typing import Dict, Any, Iterable

import pytest
import yaml

from .exceptions import TestProcessingException, DBTException
from .builtin import BUILTIN_TEST_SEQUENCES, DEFAULT_PROJECTS, DbtProject


from dbt.adapters.factory import FACTORY
from dbt.config import RuntimeConfig
from dbt.main import parse_args


class DbtSpecFile(pytest.File):
    def collect(self):
        with self.fspath.open() as fp:
            raw = yaml.safe_load(fp)
        if not raw:
            return
        try:
            raw_target = raw['target']
        except KeyError:
            raise TestProcessingException(
                'Invalid dbtspec: target not found'
            ) from None

        projects = {
            k: DbtProject.from_dict(v)
            for k, v in DEFAULT_PROJECTS.items()
        }

        for project in raw.get('projects', []):
            parsed = DbtProject.from_dict(project, projects)
            projects[parsed.name] = parsed

        try:
            sequences = raw['sequences']
        except KeyError:
            raise TestProcessingException(
                'Invalid dbtspec: sequences not found'
            ) from None

        for name, testdef in sequences.items():
            if isinstance(testdef, str):
                try:
                    testdef = BUILTIN_TEST_SEQUENCES[testdef]
                except KeyError:
                    raise TestProcessingException(
                        f'Unknown builtin test name {testdef}'
                    )
            try:
                project_name = testdef['project']
            except KeyError:
                raise TestProcessingException(
                    f'Invalid dbtspec: no project in sequence {testdef}'
                ) from None

            try:
                project = projects[project_name]
            except KeyError:
                raise TestProcessingException(
                    f'Invalid dbtspec: project {project_name} unknown'
                ) from None

            try:
                sequence = testdef['sequence']
            except KeyError:
                raise TestProcessingException(
                    f'Invalid dbtspec: no sequence in sequence {testdef}'
                ) from None

            yield DbtItem.from_parent(
                self,
                name=name,
                target=raw_target,
                sequence=sequence,
                project=project,
            )


class DbtItem(pytest.Item):
    def __init__(self, name, parent, target, sequence, project):
        super().__init__(name, parent)
        self.target = target
        self.sequence = sequence
        self.project = project
        self.adapter = None
        self.schema_relation = None
        start = datetime.utcnow().strftime('%y%m%d%H%M%S%f')
        randval = random.SystemRandom().randint(0, 999999)
        self.random_suffix = f'{start}{randval:06}'

    def _base_vars(self):
        return {'_dbt_random_suffix': self.random_suffix}

    def _get_adapter(self, tmpdir):
        project_path = os.path.join(tmpdir, 'project')
        args = parse_args([
            'compile', '--profile', 'dbt-pytest', '--target', 'default',
            '--project-dir', project_path, '--profiles-dir', tmpdir,
            '--vars', yaml.safe_dump(self._base_vars()),
        ])
        with open(os.path.join(args.profiles_dir, 'profiles.yml')) as fp:
            data = yaml.safe_load(fp)
            try:
                profile = data[args.profile]
            except KeyError:
                raise ValueError(f'profile {args.profile} not found')
            try:
                outputs = profile['outputs']
            except KeyError:
                raise ValueError(f'malformed profile {args.profile}')
            try:
                target = outputs[args.target]
            except KeyError:
                raise ValueError(
                    f'target {args.target} not found in {args.profile}'
                )
            try:
                adapter_type = target['type']
            except KeyError:
                raise ValueError(
                    f'target {args.target} in {args.profile} has no type')
        _ = FACTORY.load_plugin(adapter_type)
        config = RuntimeConfig.from_args(args)

        FACTORY.register_adapter(config)
        adapter = FACTORY.lookup_adapter(config.credentials.type)
        return adapter

    @staticmethod
    def _get_from_dict(dct: Dict[str, Any], keypath: Iterable[str]):
        value = dct
        for key in keypath:
            value = value[key]
        return value

    def _update_nested_dict(
        dct: Dict[str, Any], keypath: Iterable[str], value: Any
    ):
        next_key, keypath = keypath[0], keypath[1:]
        for cur_key in keypath:
            if next_key not in dct:
                dct[next_key] = {}
            dct = dct[next_key]
            next_key = cur_key
        dct[next_key] = value

    def get_fact(self, key):
        if isinstance(key, str) and key.startswith('fact.'):
            parts = key.split('.')[1:]
            try:
                return self._get_from_dict(self.project.facts, parts)
            except KeyError:
                pass
        return key

    def _relation_from_name(self, name: str):
        """reverse-engineer a relation (including quoting) from a given name and
        the adapter.

        This does assume that relations are split by the `.` character.

        Note that this doesn't really have to be correct, it only has to
        round-trip properly. Still, do our best to get this right.
        """
        cls = self.adapter.Relation
        credentials = self.adapter.config.credentials
        quote_policy = cls.get_default_quote_policy().to_dict()
        include_policy = cls.get_default_include_policy().to_dict()
        kwargs = {}

        parts = name.split('.')
        if len(parts) == 0:  # I think this is literally impossible!
            raise TestProcessingException(f'Invalid test name {name}')

        names = ['database', 'schema', 'identifier']
        defaults = [credentials.database, credentials.schema, None]
        values = chain(repeat(None, 3 - len(parts)), parts)
        for name, value, default in zip(names, values, defaults):
            # no quote policy -> use the default
            if value is None:
                if default is None:
                    include_policy[name] = False
                value = default
            else:
                include_policy[name] = True
                # if we have a value, we can figure out the quote policy.
                trimmed = value[1:-1]
                if self.adapter.quote(trimmed) == value:
                    quote_policy[name] = True
                    value = trimmed
                else:
                    quote_policy[name] = False
            kwargs[name] = value

        return cls.create(
            include_policy=include_policy,
            quote_policy=quote_policy,
            **kwargs
        )

    def step_dbt(self, sequence_item, tmpdir):
        if 'cmd' not in sequence_item:
            raise TestProcessingException(
                f'Got item type cmd, but no cmd in {sequence_item}'
            )
        cmd = shlex.split(sequence_item['cmd'])
        partial_parse = sequence_item.get('partial_parse', False)
        extra = [
            '--target', 'default',
            '--profile', 'dbt-pytest',
            '--profiles-dir', tmpdir,
            '--project-dir', os.path.join(tmpdir, 'project')
        ]
        base_cmd = ['dbt', '--debug']

        if partial_parse:
            base_cmd.append('--partial-parse')
        else:
            base_cmd.append('--no-partial-parse')

        full_cmd = base_cmd + cmd + extra
        cli_vars = sequence_item.get('vars', {}).copy()
        cli_vars.update(self._base_vars())
        if cli_vars:
            full_cmd.extend(('--vars', yaml.safe_dump(cli_vars)))
        expect_passes = sequence_item.get('check', True)
        result = run(full_cmd, check=False, stdout=PIPE, stderr=PIPE)
        print(result.stdout.decode('utf-8'))
        if expect_passes:
            if result.returncode != 0:
                raise TestProcessingException(
                    f'Command {full_cmd} failed, expected pass! Got '
                    f'rc={result.returncode}'
                )
        else:
            if result.returncode == 0:
                raise TestProcessingException(
                    f'Command {full_cmd} passed, expected failure! Got '
                    f'rc={result.returncode}'
                )
        return result

    @staticmethod
    def _build_expected_attributes_dict(
        values: Dict[str, Any]
    ) -> Dict[str, Any]:
        # turn keys into nested dicts
        attributes = {}
        for key, value in values.items():
            parts = key.split('.', 1)
            if len(parts) != 2:
                raise TestProcessingException(
                    f'Expected a longer keypath, only got "{key}" '
                    '(no attributes?)'
                )
            name, keypath = parts

            if name not in attributes:
                attributes[name] = {}
            attributes[name][keypath] = value
        return attributes

    @staticmethod
    def _get_name(
        result: Dict[str, Any],
        nodes: Dict[str, Any]
    ) -> str:
        """Given a run result get the unique_id and lookup the name from
        a dict of nodes mapped to their unique_id.
        """
        try:
            unique_id = result['unique_id']
        except KeyError as exc:
            raise DBTException(
                f'Invalid result, missing required key {exc}'
            ) from None
        try:
            return nodes[unique_id]['name']
        except KeyError as exc:
            raise DBTException(
                f'Invalid node, missing required key {exc}'
            ) from None

    def step_run_results(self, sequence_item, tmpdir):
        run_results_path = os.path.join(
            tmpdir, 'project', 'target', 'run_results.json')
        manifest_path = os.path.join(
            tmpdir, 'project', 'target', 'manifest.json')

        expect_exists = sequence_item.get('exists', True)

        assert expect_exists == os.path.exists(run_results_path)
        if not expect_exists:
            return None

        try:
            with open(
                run_results_path
            ) as results_fp, open(
                manifest_path
            ) as manifest_fp:
                run_results_data = json.load(results_fp)
                manifest_data = json.load(manifest_fp)
        except Exception as exc:
            raise DBTException(
                f'could not load run_results.json: {exc}'
            ) from exc
        try:
            results = run_results_data['results']
        except KeyError:
            raise DBTException(
                'Invalid run_results.json - no results'
            ) from None
        try:
            nodes = manifest_data['nodes']
        except KeyError:
            raise DBTException(
                'Invalid manifest.json - no nodes'
            ) from None

        if 'length' in sequence_item:
            expected = self.get_fact(sequence_item['length'])
            assert expected == len(results)
        if 'names' in sequence_item:
            expected_names = set(self.get_fact(sequence_item['names']))
            extra_results_ok = sequence_item.get('extra_results_ok', False)

            for result in results:
                name = self._get_name(result, nodes)
                if (not extra_results_ok) and (name not in expected_names):
                    raise DBTException(
                        f'Got unexpected name {name} in results'
                    )
                expected_names.discard(name)
            if expected_names:
                raise DBTException(
                    f'Nodes missing from run_results: {list(expected_names)}'
                )
        if 'attributes' in sequence_item:
            values = self.get_fact(sequence_item['attributes'])

            attributes = self._build_expected_attributes_dict(values)

            for result in results:
                name = self._get_name(result, nodes)
                if name in attributes:
                    for key, value in attributes[name].items():
                        try:
                            self._get_from_dict(result, key.split('.'))
                        except KeyError as exc:
                            raise DBTException(
                                f'Invalid result, missing required key {exc}'
                            ) from None

    def _expected_catalog_member(self, sequence_item, catalog, member_name):
        if member_name not in catalog:
            raise DBTException(
                f'invalid catalog.json: no {member_name}!'
            )

        actual = catalog[member_name]
        expected = sequence_item.get(member_name, {})
        if 'length' in expected:
            expected_length = self.get_fact(expected['length'])
            assert len(actual) == expected_length

        if 'names' in expected:
            extra_nodes_ok = expected.get('extra_nodes_ok', False)
            expected_names = set(self.get_fact(expected['names']))
            for node in actual.values():
                try:
                    name = node['metadata']['name']
                except KeyError as exc:
                    singular = member_name[:-1]
                    raise TestProcessingException(
                        f'Invalid catalog {singular}: missing key {exc}'
                    ) from None
                if (not extra_nodes_ok) and (name not in expected_names):
                    raise DBTException(
                        f'Got unexpected name {name} in catalog'
                    )
                expected_names.discard(name)
            if expected_names:
                raise DBTException(
                    f'{member_name.title()} missing from run_results: '
                    f'{list(expected_names)}'
                )

    def step_catalog(self, sequence_item, tmpdir):
        path = os.path.join(tmpdir, 'project', 'target', 'catalog.json')
        expect_exists = sequence_item.get('exists', True)

        assert expect_exists == os.path.exists(path)
        if not expect_exists:
            return None

        try:
            with open(path) as fp:
                catalog = json.load(fp)
        except Exception as exc:
            raise DBTException(
                f'could not load catalog.json: {exc}'
            ) from exc

        self._expected_catalog_member(sequence_item, catalog, 'nodes')
        self._expected_catalog_member(sequence_item, catalog, 'sources')

    def step_relations_equal(self, sequence_item):
        if 'relations' not in sequence_item:
            raise TestProcessingException(
                'Invalid relations_equal: no relations'
            )
        relation_names = self.get_fact(sequence_item['relations'])
        assert isinstance(relation_names, list)
        if len(relation_names) < 2:
            raise TestProcessingException(
                'Not enough relations to compare',
            )
        relations = [
            self._relation_from_name(name) for name in relation_names
        ]
        with self.adapter.connection_named('_test'):
            basis, compares = relations[0], relations[1:]
            columns = [
                c.name for c in self.adapter.get_columns_in_relation(basis)
            ]

            for relation in compares:
                sql = self.adapter.get_rows_different_sql(
                    basis, relation, column_names=columns
                )
                _, tbl = self.adapter.execute(sql, fetch=True)
                num_rows = len(tbl)
                assert num_rows == 1, f'Invalid sql query from get_rows_different_sql: incorrect number of rows ({num_rows})'
                num_cols = len(tbl[0])
                assert num_cols == 2, f'Invalid sql query from get_rows_different_sql: incorrect number of cols ({num_cols})'
                row_count_difference = tbl[0][0]
                assert row_count_difference == 0, f'Got {row_count_difference} difference in row count betwen {basis} and {relation}'
                rows_mismatched = tbl[0][1]
                assert rows_mismatched == 0, f'Got {rows_mismatched} different rows between {basis} and {relation}'

    def step_relation_rows(self, sequence_item):
        if 'name' not in sequence_item:
            raise TestProcessingException('Invalid relation_rows: no name')
        if 'length' not in sequence_item:
            raise TestProcessingException('Invalid relation_rows: no length')
        name = self.get_fact(sequence_item['name'])
        length = self.get_fact(sequence_item['length'])
        relation = self._relation_from_name(name)
        with self.adapter.connection_named('_test'):
            _, tbl = self.adapter.execute(
                f'select count(*) as num_rows from {relation}',
                fetch=True
            )

        assert len(tbl) == 1 and len(tbl[0]) == 1, \
            'count did not return 1 row with 1 column'
        assert tbl[0][0] == length, \
            f'expected {name} to have {length} rows, but it has {tbl[0][0]}'

    def _generate_update_clause(self, clause) -> str:
        if 'type' not in clause:
            raise TestProcessingException(
                'invalid update_rows clause: no type'
            )
        clause_type = clause['type']

        if clause_type == 'add_timestamp':
            if 'src_col' not in clause:
                raise TestProcessingException(
                    'Invalid update_rows clause: no src_col'
                )
            add_to = self.get_fact(clause['src_col'])
            kwargs = {
                k: self.get_fact(v) for k, v in clause.items()
                if k in ('interval', 'number')
            }
            with self.adapter.connection_named('_test'):
                return self.adapter.timestamp_add_sql(
                    add_to=add_to,
                    **kwargs
                )
        elif clause_type == 'add_string':
            if 'src_col' not in clause:
                raise TestProcessingException(
                    'Invalid update_rows clause: no src_col'
                )
            if 'value' not in clause:
                raise TestProcessingException(
                    'Invalid update_rows clause: no value'
                )
            src_col = self.get_fact(clause['src_col'])
            value = self.get_fact(clause['value'])
            location = clause.get('location', 'append')
            with self.adapter.connection_named('_test'):
                return self.adapter.string_add_sql(
                    src_col, value, location
                )
        else:
            raise TestProcessingException(
                f'Unknown clause type in update_rows: {clause_type}'
            )

    def step_relation_types(self, sequence_item):
        """
        type: relation_types
        expect:
            foo: view
            bar: table
        """
        if 'expect' not in sequence_item:
            raise TestProcessingException('Invalid relation_types: no expect')
        expected = self.get_fact(sequence_item['expect'])

        expected_relation_values = {}
        found_relations = []
        schemas = set()

        for key, value in expected.items():
            relation = self._relation_from_name(key)
            expected_relation_values[relation] = value
            schemas.add(relation.without_identifier())
        with self.adapter.connection_named('__test'):
            for schema in schemas:
                found_relations.extend(
                    self.adapter.list_relations_without_caching(schema))

        for key, value in expected.items():
            for relation in found_relations:
                # this might be too broad
                if relation.identifier == key:
                    assert relation.type == value, (
                        f'Got an unexpected relation type of {relation.type} '
                        f'for relation {key}, expected {value}'
                    )

    def step_update_rows(self, sequence_item):
        """
            type: update_rows
            name: base
            dst_col: some_date
            clause:
              type: add_timestamp
              src_col: some_date
            where: id > 10
        """
        if 'name' not in sequence_item:
            raise TestProcessingException('Invalid update_rows: no name')
        if 'dst_col' not in sequence_item:
            raise TestProcessingException('Invalid update_rows: no dst_col')

        if 'clause' not in sequence_item:
            raise TestProcessingException('Invalid update_rows: no clause')

        clause = self.get_fact(sequence_item['clause'])
        if isinstance(clause, dict):
            clause = self._generate_update_clause(clause)

        where = None
        if 'where' in sequence_item:
            where = self.get_fact(sequence_item['where'])

        name = self.get_fact(sequence_item['name'])
        dst_col = self.get_fact(sequence_item['dst_col'])
        relation = self._relation_from_name(name)

        with self.adapter.connection_named('_test'):
            sql = self.adapter.update_column_sql(
                dst_name=str(relation),
                dst_column=dst_col,
                clause=clause,
                where_clause=where,
            )
            self.adapter.execute(sql, auto_begin=True)
            self.adapter.commit_if_has_connection()

    def _write_profile(self, tmpdir):
        profile_data = {
            'config': {
                'send_anonymous_usage_stats': False,
            },
            'dbt-pytest': {
                'target': 'default',
                'outputs': {
                    'default': self.target,
                },
            },
        }
        with open(os.path.join(tmpdir, 'profiles.yml'), 'w') as fp:
            fp.write(yaml.safe_dump(profile_data))

    def _add_context(self, error_str, idx, test_item):
        item_type = test_item['type']
        return f'{error_str} in test index {idx} (item_type={item_type})'

    def run_test_item(self, idx, test_item, tmpdir):
        try:
            item_type = test_item['type']
        except KeyError:
            raise TestProcessingException(
                f'Could not find type in {test_item}'
            ) from None
        print(f'Executing step {idx+1}/{len(self.sequence)}')
        try:
            if item_type == 'dbt':
                assert os.path.exists(tmpdir)
                self.step_dbt(test_item, tmpdir)
            elif item_type == 'run_results':
                self.step_run_results(test_item, tmpdir)
            elif item_type == 'catalog':
                self.step_catalog(test_item, tmpdir)
            elif item_type == 'relations_equal':
                self.step_relations_equal(test_item)
            elif item_type == 'relation_rows':
                self.step_relation_rows(test_item)
            elif item_type == 'update_rows':
                self.step_update_rows(test_item)
            elif item_type == 'relation_types':
                self.step_relation_types(test_item)
            else:
                raise TestProcessingException(
                    f'Unknown item type {item_type}'
                )
        except AssertionError as exc:
            if len(exc.args) == 1:
                arg = self._add_context(exc.args[0], idx, test_item)
                exc.args = (arg,)
            else:  # uhhhhhhh
                exc.args = exc.args + (self._add_context('', idx, test_item),)
            raise

    def runtest(self):
        FACTORY.reset_adapters()
        with tempfile.TemporaryDirectory() as tmpdir:
            self._write_profile(tmpdir)
            self.project.write(tmpdir)
            self.adapter = self._get_adapter(tmpdir)

            self.schema_relation = self.adapter.Relation.create(
                database=self.adapter.config.credentials.database,
                schema=self.adapter.config.credentials.schema,
                quote_policy=self.adapter.config.quoting,
            )

            try:
                for idx, test_item in enumerate(self.sequence):
                    self.run_test_item(idx, test_item, tmpdir)
            finally:
                with self.adapter.connection_named('__test'):
                    if self.config.getoption('drop_schema'):
                        self.adapter.drop_schema(self.schema_relation)

        return True

    def repr_failure(self, excinfo):
        """ called when self.runtest() raises an exception. """
        if isinstance(excinfo.value, DBTException):
            return "\n".join([
                "usecase execution failed",
                "   spec failed: {!r}".format(excinfo.value.args),
                "   no further details known at this point.",
            ])
        elif isinstance(excinfo.value, CalledProcessError):
            failed = str(excinfo.value.cmd)
            stdout = excinfo.value.stdout.decode('utf-8')
            stderr = excinfo.value.stderr.decode('utf-8')
            return '\n'.join([
                f'failed to execute "{failed}:',
                f'   output: {stdout}',
                f'   error: {stderr}',
                f'   rc: {excinfo.value.returncode}',
            ])
        elif isinstance(excinfo.value, TestProcessingException):
            return str(excinfo.value)
        else:
            return f'Unknown error: {excinfo.value}'

    def reportinfo(self):
        return self.fspath, 0, "usecase: {}".format(self.name)



from setuptools import setup


def read(path):
    with open(path, encoding='utf-8') as fp:
        return fp.read()


setup(
    name='pytest-dbt-adapter',
    packages=['pytest_dbt_adapter', 'dbt_adapter_tests'],
    author="Fishtown Analytics",
    author_email="info@fishtownanalytics.com",
    url="https://github.com/fishtown-analytics/dbt-adapter-tests",
    version='0.5.1',
    package_data={
        'pytest_dbt_adapter': [
            'projects/*.yml',
            'sequences/*.yml',
        ]
    },
    entry_points={
        'pytest11': [
            'pytest_dbt_adapter = pytest_dbt_adapter',
        ]
    },
    install_requires=['py>=1.3.0', 'pytest>=6,<7', 'pyyaml>=3.11'],
    description="A pytest plugin for testing dbt adapter plugins",
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    python_requires=">=3.6.2",
)
